#include <cstring>
#include <cctype>
#include <cstdlib>
#include <cstdio>
#include <assert.h>
#include <unistd.h>

using namespace std;

typedef unsigned long long postype;

static unsigned sectorsize = 512;

#include <unistd.h>
#include <linux/unistd.h>   /* for __NR__llseek */

static _syscall5(int, _llseek, uint, fd, ulong, hi, ulong, lo, loff_t *, res, uint, wh);
static postype my_llseek(int fd, postype offset, unsigned origin)
{
    postype result=0;
    int retval;

    retval = _llseek (fd,
            ((postype) offset) >> 32LL,
            ((postype) offset) & 0xffffffffUL,
            (loff_t *)&result, origin);
    return (retval == -1 ? (postype) retval : result);
}

static postype getbufsize(const char *s)
{
     postype value=0;
     while(isdigit(*s))
          value=value*10 + (*s++ - '0');
     if(!strcmp("c", s))value*=1;
     else if(!strcmp("w", s))value*=2;
     else if(!strcmp("b", s))value*=512;
     else if(!strcmp("kD", s))value *= 1000;
     else if(!strcmp("k", s))value *= 1024;
     else if(!strcmp("MD", s))value *= 1000*1000;
     else if(!strcmp("M", s))value *= 1024*1024;
     else if(!strcmp("GD", s))value *= 1000*1000*1000;
     else if(!strcmp("G", s))value *= 1024LL*1024*1024;
     else if(!strcmp("TD", s))value *= 1000LL*1000*1000*1000;
     else if(!strcmp("T", s))value *= 1024LL*1024*1024*1024;
     else if(!strcmp("PD", s))value *= 1000LL*1000*1000*1000*1000;
     else if(!strcmp("P", s))value *= 1024LL*1024*1024*1024*1024;
     else if(!strcmp("ED", s))value *= 1000LL*1000*1000*1000*1000*1000;
     else if(!strcmp("E", s))value *= 1024LL*1024*1024*1024*1024*1024;
     else if(!strcmp("ZD", s))value *= 1000LL*1000*1000*1000*1000*1000*1000;
     else if(!strcmp("Z", s))value *= 1024LL*1024*1024*1024*1024*1024*1024;
     else if(!strcmp("YD", s))value *= 1000LL*1000*1000*1000*1000*1000*1000*1000;
     else if(!strcmp("Y", s))value *= 1024LL*1024*1024*1024*1024*1024*1024*1024;
     else if(*s)
     {
          fprintf(stderr, "Invalid size suffix: '%s'\n", s);
          exit(1);
     }
     return value;
}

/* Write a sector at a time */
static void CarefulWrite(postype pos, int fd,
                         char *buf, postype bufsize)
{
    my_llseek(fd, pos, SEEK_SET);
    while(bufsize > 0)
    {
        postype sectorbegin = pos & ~(sectorsize-1);
        postype nextsectorbegin = sectorbegin + sectorsize;
        unsigned bytes = nextsectorbegin-pos;
        
        if(bytes > bufsize)
            bytes = bufsize;
        
        assert(bytes != 0);
        
        int res = write(fd, buf, bytes);
        if(res == bytes)
        {
            pos += bytes;
            buf += bytes;
            bufsize -= bytes;
            continue;
        }
        char ErrBuf[256];
        sprintf(ErrBuf, "write[sector %lld]", sectorbegin / sectorsize);
        perror(ErrBuf);
        
        pos += bytes;
        buf += bytes;
        bufsize -= bytes;
        
        my_llseek(fd, pos, SEEK_SET);
    }
}

static void WriteBytes(int fd, char *buf, postype bytecount, postype pos)
{
    int res2 = write(fd, buf, bytecount);
    if(res2 < 0)
    {
        char ErrBuf[256];
        sprintf(ErrBuf, "write[@%lld,%llu bytes]", pos, bytecount);
        perror(ErrBuf);

        CarefulWrite(pos, fd, buf, bytecount);
    }
}

/* Read a sector at a time */
static void CarefulCopy(postype inpos, int infd,
                        postype outpos, int outfd,
                        char *buf, postype bufsize)
{
    my_llseek(infd, inpos, SEEK_SET);
    while(bufsize > 0)
    {
        postype sectorbegin = inpos & ~(sectorsize-1);
        postype nextsectorbegin = sectorbegin + sectorsize;
        unsigned bytes = nextsectorbegin-inpos;
        
        if(bytes > bufsize)
            bytes = bufsize;
        
        assert(bytes != 0);
        
        int res = read(infd, buf, bytes);
        if(res > 0)
        {
            WriteBytes(outfd, buf, res, outpos);
            outpos += res;
            inpos += res;
            bufsize -= res;
            continue;
        }
        char ErrBuf[256];
        sprintf(ErrBuf, "read[sector %lld]", sectorbegin / sectorsize);
        perror(ErrBuf);
        
        outpos += bytes;
        inpos += bytes;
        bufsize -= bytes;
        
        my_llseek(infd, inpos, SEEK_SET);
        my_llseek(outfd, outpos, SEEK_SET);
    }
}

static bool EOFreached = false;

static void CopyData(int infd, postype inputptr,
                     int outfd, postype outputptr,
                     char *buf, postype bufsize)
{
Continue: ;
    int res = read(infd, buf, bufsize);
    if(res > 0)
    {
        WriteBytes(outfd, buf, res, outputptr);
NextRead:
        outputptr += res;
        inputptr += res;
        bufsize -= res;
        if(bufsize)
        {
            fprintf(stderr, "short read...\n");
            goto Continue;
        }
    }
    else if(res < 0)
    {
        char ErrBuf[256];
        sprintf(ErrBuf, "read[@%lld,%llu bytes]", inputptr, bufsize);
        perror(ErrBuf);
        
        CarefulCopy(inputptr, infd,
                    outputptr, outfd,
                    buf, bufsize);
        res = bufsize;
        goto NextRead;
    }
    else
    {
        EOFreached = true;
    }
}

int main(int argc, const char *const *argv)
{
     // default settings:
     const char *infn = NULL;
     const char *outfn = NULL;
     postype inbufsize = 1024;
     postype outbufsize = 1024;
     postype inputskip = 0;
     postype outputskip = 0;
     postype blockcount = 0;
     
     for(int c=1; c<argc; ++c)
     {
          // if=filename
          // of=filename
          // bs=size
          // ibs=size
          // obs=size
          // skip=iblocks
          // seek=oblocks
          // count=iblocks
          
          const char *s = argv[c];
          if(!strncmp(s, "if=", 3))
              infn = s+3;
          else if(!strncmp(s, "of=", 3))
              outfn = s+3;
          else if(!strncmp(s, "bs=", 3))
              inbufsize = outbufsize = getbufsize(s+3);
          else if(!strncmp(s, "ibs=", 4))
              inbufsize = getbufsize(s+4);
          else if(!strncmp(s, "obs=", 4))
              outbufsize = getbufsize(s+4);
          else if(!strncmp(s, "skip=", 5))
              inputskip = getbufsize(s+5);
          else if(!strncmp(s, "seek=", 5))
              outputskip = getbufsize(s+5);
          else if(!strncmp(s, "count=", 6))
              blockcount = getbufsize(s+6);
          else if(!strcmp(s, "--help"))
          {
              printf("Failsafe dd v"VERSION" - Copyright (C) 1992,2002 Bisqwit (http://bisqwit.iki.fi/)\n"
                     "faildd comes under GPL and is purposely made to feel like the standard dd.\n"
                     "Usage:\n"
                     "   faildd [OPTION] ...\n"
                     "Copy a file, not choking when disk errors happen.\n"
                     "\n"
                     "  bs=BYTES        force ibs=BYTES and obs=BYTES\n"
                     "  count=BLOCKS    copy only BLOCKS input blocks\n"
                     "  ibs=BYTES       read BYTES bytes at a time\n"
                     "  if=FILE         read from FILE instead of stdin\n"
                     "  obs=BYTES       write BYTES bytes at a time\n"
                     "  of=FILE         write to FILE instead of stdout\n"
                     "  seek=BLOCKS     skip BLOCKS obs-sized blocks at start of output\n"
                     "  skip=BLOCKS     skip BLOCKS ibs-sized blocks at start of input\n"
                     "      --help      display this help and exit\n"
                     "      --version   output version information and exit\n"
                     "\n"
                     "BLOCKS and BYTES may be followed by the following multiplicative suffixes:\n"
                     "xM M, c 1, w 2, b 512, kD 1000, k 1024, MD 1,000,000, M 1,048,576,\n"
                     "GD 1,000,000,000, G 1,073,741,824, and so on for T, P, E, Z, Y.\n"
                  );

              return 0;
          }
          else if(!strcmp(s, "--version"))
          {
              printf("Failsafe dd v"VERSION" - Copyright (C) 1992,2002 Bisqwit (http://bisqwit.iki.fi/)\n");
              return 0;
          }
          else
          {
              fprintf(stderr, "dd: Unrecognized option `%s'\n", s);
              return 4;
          }
     }
     inputskip  *= inbufsize;
     outputskip *= outbufsize;
     
     if(inbufsize != outbufsize)
     {
          outbufsize = inbufsize;
          fprintf(stderr, "Warning: Input buffer size and output buffer size differ.\n"
                          "         This program writes as much as it reads.\n"
                          "         Ignoring output buffer size.\n");
     }
     
     if(infn)
     {
          bool fail = false;
          if(!freopen(infn, "rb", stdin))
          { perror(infn); fail = true; }
          if(!freopen(outfn, "wb", stdout))
          { perror(outfn); fail = true; }
          if(fail)
               exit(2);
     }
     
     int infd = fileno(stdin);
     int outfd = fileno(stdout);
     
     if(inputskip)my_llseek(infd, inputskip, SEEK_SET);
     if(outputskip)my_llseek(outfd, outputskip, SEEK_SET);
     
     if(sectorsize & (sectorsize-1))
     {
         fprintf(stderr, "Error: sector size %u is not a power of two.\n", sectorsize);
         return 3;
     }
     
     char *inbuf = new char[inbufsize];
     postype blockpos=0;
     for(;;)
     {
     	 fprintf(stderr, "%14llu/%-14llu\r", blockpos, blockcount);
     	 
         CopyData(infd,inputskip,
                  outfd,outputskip,
                  inbuf, inbufsize);
         if(blockcount)
         {
             --blockcount;
             if(!blockcount)break;
         }
         inputskip += inbufsize;
         outputskip += outbufsize;
         
         if(EOFreached)
         {
             if(blockcount)
                 fprintf(stderr, "Premature EOF reached\n");
             break;
         }
         ++blockpos;
     }
     delete[] inbuf;
     return 0;
}
